//#####################################################################
// Copyright 2010, Wen Zheng.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifdef USE_OSMESA

#include <PhysBAM_Tools/Log/LOG.h>
#include <PhysBAM_Rendering/PhysBAM_OpenGL/OpenGL/OPENGL_OSMESA_PBUFFER.h>
#include <cstdio>
#include <cstring>
#include <iostream>
using namespace PhysBAM;

OPENGL_PBUFFER::OPENGL_PBUFFER()
    :verbose(false),buffer(0)
{LOG::cout<<"OSMESA"<<std::endl;}

OPENGL_PBUFFER::~OPENGL_PBUFFER()
{
    Destroy();
}

bool OPENGL_PBUFFER::
Create(int width, int height)
{
#if OSMESA_MAJOR_VERSION * 100 + OSMESA_MINOR_VERSION >= 305
    /* specify Z, stencil, accum sizes */
    ctx = OSMesaCreateContextExt( OSMESA_RGBA, 16, 0, 0, NULL );
#else
    ctx = OSMesaCreateContext( OSMESA_RGBA, NULL );
#endif
    if (!ctx) {
        LOG::cerr<<"OSMesaCreateContext failed!"<<std::endl;
        return false;}

    buffer = malloc( width * height * 4 * sizeof(GLubyte));
    if (!buffer) {
        LOG::cerr<<"Alloc image buffer failed!"<<std::endl;
        return false;}

    if (!OSMesaMakeCurrent( ctx, buffer, GL_UNSIGNED_BYTE, width, height )) {
        LOG::cerr<<"OSMesaMakeCurrent failed!"<<std::endl;
        return false;}

    return true;
}

void OPENGL_PBUFFER::
Destroy()
{
    if(buffer) free(buffer);
    OSMesaDestroyContext(ctx);
}

#endif
