//#####################################################################
// Copyright 2003-2008, Eran Guendelman, Sergey Koltakov, Michael Lentine, Andrew Selle, Tamar Shinar, Eftychios Sifakis.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Rendering/PhysBAM_OpenGL/OpenGL/OPENGL_SELECTION.h>
using namespace PhysBAM;
OPENGL_SELECTION::
OPENGL_SELECTION(TYPE type,OPENGL_OBJECT* object)
    :type(type),object(object),hide(false)
{
}
OPENGL_SELECTION::
~OPENGL_SELECTION()
{
}
OPENGL_SELECTION::TYPE OPENGL_SELECTION::
Actual_Type() const
{
    return type;
}
template<class TV> void OPENGL_SELECTION::
Draw_Highlighted_Vertex(const TV& position,int id,const OPENGL_COLOR& color)
{
    glPushAttrib(GL_POINT_BIT | GL_ENABLE_BIT | GL_CURRENT_BIT);
    glDisable(GL_LIGHTING);
    glPointSize(OPENGL_PREFERENCES::highlighted_point_size);
    color.Send_To_GL_Pipeline();
    OpenGL_Begin(GL_POINTS);
    OpenGL_Vertex(position);
    OpenGL_End();
    if(id) OpenGL_String(position,STRING_UTILITIES::string_sprintf("   %d",id));
    glPopAttrib();
}
template<class T> void OPENGL_SELECTION::
Draw_Highlighted_Quad(const VECTOR<T,2>& x00,const VECTOR<T,2>& x11,const OPENGL_COLOR& color)
{
    glPushAttrib(GL_LINE_BIT | GL_ENABLE_BIT | GL_CURRENT_BIT);
    glDisable(GL_LIGHTING);
    glLineWidth(OPENGL_PREFERENCES::highlighted_line_width);
    color.Send_To_GL_Pipeline();
    OpenGL_Begin(GL_LINE_LOOP);
    OpenGL_Quad_2D(x00,x11);
    OpenGL_End();
    glPopAttrib();
}
template<class T> void OPENGL_SELECTION::
Draw_Highlighted_Quad(const VECTOR<T,3>& node1,const VECTOR<T,3>& node2,const VECTOR<T,3>& node3,const VECTOR<T,3>& node4,const OPENGL_COLOR& color)
{
    glPushAttrib(GL_LINE_BIT | GL_ENABLE_BIT | GL_CURRENT_BIT);
    glDisable(GL_LIGHTING);
    glLineWidth(OPENGL_PREFERENCES::highlighted_line_width);
    color.Send_To_GL_Pipeline();
    OpenGL_Begin(GL_LINE_LOOP);
    OpenGL_Vertex(node1);OpenGL_Vertex(node2);OpenGL_Vertex(node4);OpenGL_Vertex(node3);
    OpenGL_End();
    glPopAttrib();
}
template<class T> void OPENGL_SELECTION::
Draw_Highlighted_Box(const VECTOR<T,3>& x000,const VECTOR<T,3>& x111,const OPENGL_COLOR& color)
{
    glPushAttrib(GL_LINE_BIT | GL_ENABLE_BIT | GL_CURRENT_BIT);
    glDisable(GL_LIGHTING);
    glLineWidth(OPENGL_PREFERENCES::highlighted_line_width);
    color.Send_To_GL_Pipeline();
    
    VECTOR<T,3> x001(x000.x,x000.y,x111.z),x010(x000.x,x111.y,x000.z),x011(x000.x,x111.y,x111.z),
        x100(x111.x,x000.y,x000.z),x101(x111.x,x000.y,x111.z),x110(x111.x,x111.y,x000.z);
    
    OpenGL_Begin(GL_LINE_LOOP);
    OpenGL_Vertex(x000); OpenGL_Vertex(x010); OpenGL_Vertex(x011); OpenGL_Vertex(x001);
    OpenGL_End();
    OpenGL_Begin(GL_LINE_LOOP);
    OpenGL_Vertex(x100); OpenGL_Vertex(x110); OpenGL_Vertex(x111); OpenGL_Vertex(x101);
    OpenGL_End();
    OpenGL_Begin(GL_LINES);
    OpenGL_Vertex(x000); OpenGL_Vertex(x100);
    OpenGL_Vertex(x010); OpenGL_Vertex(x110);
    OpenGL_Vertex(x011); OpenGL_Vertex(x111);
    OpenGL_Vertex(x001); OpenGL_Vertex(x101);
    OpenGL_End();
    glPopAttrib();
}
template void OPENGL_SELECTION::Draw_Highlighted_Vertex<VECTOR<float,3> >(VECTOR<float,3> const&,int,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Quad<float>(VECTOR<float,2> const&,VECTOR<float,2> const&,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Vertex<VECTOR<float,2> >(VECTOR<float,2> const&,int,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Vertex<VECTOR<float,1> >(VECTOR<float,1> const&,int,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Box<float>(VECTOR<float,3> const&,VECTOR<float,3> const&,OPENGL_COLOR const&);
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template void OPENGL_SELECTION::Draw_Highlighted_Vertex<VECTOR<double,3> >(VECTOR<double,3> const&,int,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Quad<double>(VECTOR<double,2> const&,VECTOR<double,2> const&,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Vertex<VECTOR<double,2> >(VECTOR<double,2> const&,int,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Vertex<VECTOR<double,1> >(VECTOR<double,1> const&,int,OPENGL_COLOR const&);
template void OPENGL_SELECTION::Draw_Highlighted_Box<double>(VECTOR<double,3> const&,VECTOR<double,3> const&,OPENGL_COLOR const&);
#endif
