//#####################################################################
// Copyright 2003-2008, Eran Guendelman, Sergey Koltakov, Michael Lentine, Andrew Selle, Tamar Shinar, Eftychios Sifakis.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class OPENGL_SELECTION
//##################################################################### 
#ifndef __OPENGL_SELECTION__
#define __OPENGL_SELECTION__

#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Math_Tools/constants.h>
#include <PhysBAM_Tools/Math_Tools/RANGE.h>
#include <PhysBAM_Rendering/PhysBAM_OpenGL/OpenGL/OPENGL_COLOR.h>
#include <PhysBAM_Rendering/PhysBAM_OpenGL/OpenGL/OPENGL_OBJECT.h>
#include <PhysBAM_Rendering/PhysBAM_OpenGL/OpenGL/OPENGL_PREFERENCES.h>
#include <PhysBAM_Rendering/PhysBAM_OpenGL/OpenGL/OPENGL_PRIMITIVES.h>
namespace PhysBAM{

class OPENGL_SELECTION:public NONCOPYABLE
{
public:
    enum TYPE { GRID_CELL_1D,GRID_CELL_2D, GRID_NODE_2D, GRID_CELL_3D, GRID_CELL_LIST_3D, GRID_NODE_3D, GRID_NODE_LIST_3D, RLE_CELL_2D, RLE_CELL_3D, BINTREE_CELL, BINTREE_NODE, QUADTREE_CELL, QUADTREE_NODE, OCTREE_NODE, OCTREE_CELL, 
                OCTREE_FACE, POINTS_2D, POINTS_3D, COMPONENT_PARTICLES_2D, COMPONENT_PARTICLES_3D};

    TYPE type;
    OPENGL_OBJECT *object;
    float min_depth, max_depth;
    bool hide;

    OPENGL_SELECTION(TYPE type,OPENGL_OBJECT* object);
    virtual ~OPENGL_SELECTION();
    virtual RANGE<VECTOR<float,3> > Bounding_Box() const=0;
    virtual TYPE Actual_Type() const;

//#####################################################################
// Useful functions to draw highlighted primitives
//#####################################################################
    template<class TV> static void Draw_Highlighted_Vertex(const TV& position,int id=0,const OPENGL_COLOR& color=OPENGL_PREFERENCES::selection_highlight_color);
    template<class T> static void Draw_Highlighted_Quad(const VECTOR<T,2>& x00,const VECTOR<T,2>& x11,const OPENGL_COLOR& color=OPENGL_PREFERENCES::selection_highlight_color);
    template<class T> static void Draw_Highlighted_Quad(const VECTOR<T,3>& node1,const VECTOR<T,3>& node2,const VECTOR<T,3>& node3,const VECTOR<T,3>& node4,
        const OPENGL_COLOR& color=OPENGL_PREFERENCES::selection_highlight_color);
    template<class T> static void Draw_Highlighted_Box(const VECTOR<T,3>& x000,const VECTOR<T,3>& x111,const OPENGL_COLOR& color=OPENGL_PREFERENCES::selection_highlight_color);
};

}

#endif
